# Copyright (C) 2014-2020. Ben Pruitt & Nick Conway; Wyss Institute
# See LICENSE for full GPLv2 license.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
'''
tests.test_thermoanalysis
~~~~~~~~~~~~~~~~~~~~~~~~~

Unit tests for the primer3-py low level thermodynamic calculation bindings.

'''

from __future__ import print_function

import random
import sys
import unittest
from time import sleep

try:
    import resource
except (ImportError, ModuleNotFoundError):  # For Windows compatibility
    resource = None  # type: ignore

from primer3 import bindings

from . import wrappers


def _get_mem_usage():
    ''' Get current process memory usage in bytes '''
    return resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024


class TestLowLevelBindings(unittest.TestCase):

    def randArgs(self):
        self.seq1 = ''.join([
            random.choice('ATGC') for _ in
            range(random.randint(20, 59))
        ])
        self.seq2 = ''.join([
            random.choice('ATGC') for _ in
            range(random.randint(20, 59))
        ])

        self.mv_conc = random.uniform(1, 200)
        self.dv_conc = random.uniform(0, 40)
        self.dntp_conc = random.uniform(0, 20)
        self.dna_conc = random.uniform(0, 200)
        self.temp_c = random.randint(10, 70)
        self.max_loop = random.randint(10, 30)
        self.dmso_conc = 0.0
        self.dmso_fact = random.uniform(0.5, 0.7)
        self.formamide_conc = random.uniform(0.8, 1.0)
        self.annealing_temp_c = -10.0  # see oligotm_main.c

        # NOTE: commented out but useful for figuring out bugs in code
        # if they exist
        # self.seq1 = 'GTGTCCAATCGACATTTAGGAGCTACGCGGCCAGGGGCAAA'
        # self.seq2 = 'TGCATTGAATGCGTGTCACGTTATGCACGC'

        # self.mv_conc = 1.1
        # self.dv_conc = 20.0
        # self.dntp_conc = 10.0
        # self.dna_conc = 20.0
        # self.temp_c = 50.0
        # self.max_loop = 20
        # self.dmso_conc = 0.0
        # self.dmso_fact= 0.6
        # self.formamide_conc = 0.0
        # self.annealing_temp_c = -10.0  # see oligotm_main.c
        # self.max_nn_length = 60

    def test_calc_tm(self):
        '''Test basic calc_tm input'''
        for _ in range(100):
            self.randArgs()
            binding_tm = bindings.calc_tm(
                seq=self.seq1,
                mv_conc=self.mv_conc,
                dv_conc=self.dv_conc,
                dntp_conc=self.dntp_conc,
                dna_conc=self.dna_conc,
                dmso_conc=self.dmso_conc,
                dmso_fact=self.dmso_fact,
                formamide_conc=self.formamide_conc,
                annealing_temp_c=self.annealing_temp_c,
            )
            wrapper_tm = wrappers.calc_tm(
                seq=self.seq1,
                mv_conc=self.mv_conc,
                dv_conc=self.dv_conc,
                dntp_conc=self.dntp_conc,
                dna_conc=self.dna_conc,
                dmso_conc=self.dmso_conc,
                dmso_fact=self.dmso_fact,
                formamide_conc=self.formamide_conc,
                annealing_temp_c=self.annealing_temp_c,
            )
            self.assertAlmostEqual(binding_tm, wrapper_tm, delta=0.5)

    def test_calc_hairpin(self):
        '''Test basic hairpin input'''
        for _ in range(1):
            self.randArgs()
            binding_res = bindings.calc_hairpin(
                seq=self.seq1,
                mv_conc=self.mv_conc,
                dv_conc=self.dv_conc,
                dntp_conc=self.dntp_conc,
                dna_conc=self.dna_conc,
                temp_c=self.temp_c,
                max_loop=self.max_loop,
                output_structure=True,
            )
            wrapper_res = wrappers.calc_hairpin(
                seq=self.seq1,
                mv_conc=self.mv_conc,
                dv_conc=self.dv_conc,
                dntp_conc=self.dntp_conc,
                dna_conc=self.dna_conc,
                temp_c=self.temp_c,
                max_loop=self.max_loop,
            )
            self.assertAlmostEqual(binding_res.tm, wrapper_res.tm, 2)
            print(binding_res.ascii_structure)
            print()
            print(wrapper_res.ascii_structure)
            self.assertEqual(
                binding_res.ascii_structure,
                # Replace line endings for windows compat
                wrapper_res.ascii_structure.replace('\r\n', '\n'),
            )

    def test_calc_homodimer(self):
        '''Test basic homodimer input'''
        for _ in range(100):
            self.randArgs()
            binding_res = bindings.calc_homodimer(
                seq=self.seq1,
                mv_conc=self.mv_conc,
                dv_conc=self.dv_conc,
                dntp_conc=self.dntp_conc,
                dna_conc=self.dna_conc,
                temp_c=self.temp_c,
                max_loop=self.max_loop,
                output_structure=True,
            )
            wrapper_res = wrappers.calc_homodimer(
                seq=self.seq1,
                mv_conc=self.mv_conc,
                dv_conc=self.dv_conc,
                dntp_conc=self.dntp_conc,
                dna_conc=self.dna_conc,
                temp_c=self.temp_c,
                max_loop=self.max_loop,
            )
            self.assertAlmostEqual(binding_res.tm, wrapper_res.tm, 2)

            print(binding_res.ascii_structure)
            print()
            print(wrapper_res.ascii_structure)

            self.assertEqual(
                binding_res.ascii_structure,
                # Replace line endings for windows compat
                wrapper_res.ascii_structure.replace('\r\n', '\n'),
            )

    def test_calc_heterodimer(self):
        '''Test basic heterodimer input'''
        for _ in range(100):
            self.randArgs()
            binding_res = bindings.calc_heterodimer(
                seq1=self.seq1,
                seq2=self.seq2,
                mv_conc=self.mv_conc,
                dv_conc=self.dv_conc,
                dntp_conc=self.dntp_conc,
                dna_conc=self.dna_conc,
                temp_c=self.temp_c,
                max_loop=self.max_loop,
                output_structure=True,
            )
            wrapper_res = wrappers.calc_heterodimer(
                seq1=self.seq1,
                seq2=self.seq2,
                mv_conc=self.mv_conc,
                dv_conc=self.dv_conc,
                dntp_conc=self.dntp_conc,
                dna_conc=self.dna_conc,
                temp_c=self.temp_c,
                max_loop=self.max_loop,
            )
            self.assertAlmostEqual(binding_res.tm, wrapper_res.tm, 2)

            print(binding_res.ascii_structure)
            print()
            print(wrapper_res.ascii_structure)

            self.assertEqual(
                binding_res.ascii_structure,
                # Replace line endings for windows compat
                wrapper_res.ascii_structure.replace('\r\n', '\n'),
            )
            # Ensure that order of sequences does not matter
            binding_12_res = bindings.calc_heterodimer(
                seq1=self.seq1,
                seq2=self.seq2,
                mv_conc=self.mv_conc,
                dv_conc=self.dv_conc,
                dntp_conc=self.dntp_conc,
                dna_conc=self.dna_conc,
                temp_c=self.temp_c,
                max_loop=self.max_loop,
                output_structure=True,
            )
            binding_21_res = bindings.calc_heterodimer(
                seq1=self.seq2,
                seq2=self.seq1,
                mv_conc=self.mv_conc,
                dv_conc=self.dv_conc,
                dntp_conc=self.dntp_conc,
                dna_conc=self.dna_conc,
                temp_c=self.temp_c,
                max_loop=self.max_loop,
            )
            self.assertAlmostEqual(binding_12_res.tm, binding_21_res.tm, 2)

    def test_max_length_heterodimer(self):
        '''Test longest heterodimer input of 10000 mer per `THAL_MAX_SEQ` '''
        self.randArgs()

        seq_small = ''.join([
            random.choice('ATGC') for _ in
            range(59)  # max size for sequence
        ])
        seq_big = ''.join([
            random.choice('ATGC') for _ in
            range(10000)  # max size for sequence 2
        ])
        binding_res = bindings.calc_heterodimer(
            seq1=seq_small,
            seq2=seq_big,
            mv_conc=self.mv_conc,
            dv_conc=self.dv_conc,
            dntp_conc=self.dntp_conc,
            dna_conc=self.dna_conc,
            temp_c=self.temp_c,
            max_loop=self.max_loop,
            output_structure=True,
        )
        wrapper_res = wrappers.calc_heterodimer(
            seq1=seq_small,
            seq2=seq_big,
            mv_conc=self.mv_conc,
            dv_conc=self.dv_conc,
            dntp_conc=self.dntp_conc,
            dna_conc=self.dna_conc,
            temp_c=self.temp_c,
            max_loop=self.max_loop,
        )
        self.assertAlmostEqual(binding_res.tm, wrapper_res.tm, 2)

        print(binding_res.ascii_structure)
        print()
        print(wrapper_res.ascii_structure)

        self.assertEqual(
            binding_res.ascii_structure,
            # Replace line endings for windows compat
            wrapper_res.ascii_structure.replace('\r\n', '\n'),
        )
        # Ensure that order of sequences does not matter
        binding_12_res = bindings.calc_heterodimer(
            seq1=seq_small,
            seq2=seq_big,
            mv_conc=self.mv_conc,
            dv_conc=self.dv_conc,
            dntp_conc=self.dntp_conc,
            dna_conc=self.dna_conc,
            temp_c=self.temp_c,
            max_loop=self.max_loop,
            output_structure=True,
        )
        binding_21_res = bindings.calc_heterodimer(
            seq1=seq_big,
            seq2=seq_small,
            mv_conc=self.mv_conc,
            dv_conc=self.dv_conc,
            dntp_conc=self.dntp_conc,
            dna_conc=self.dna_conc,
            temp_c=self.temp_c,
            max_loop=self.max_loop,
        )
        self.assertAlmostEqual(binding_12_res.tm, binding_21_res.tm, 2)

    def test_calc_end_stability(self):
        '''Test calc_end_stability'''
        for _ in range(100):
            self.randArgs()
            binding_res = bindings.calc_end_stability(
                seq1=self.seq1,
                seq2=self.seq2,
                mv_conc=self.mv_conc,
                dv_conc=self.dv_conc,
                dntp_conc=self.dntp_conc,
                dna_conc=self.dna_conc,
                temp_c=self.temp_c,
                max_loop=self.max_loop,
            )
            wrapper_res = wrappers.calc_end_stability(
                seq1=self.seq1,
                seq2=self.seq2,
                mv_conc=self.mv_conc,
                dv_conc=self.dv_conc,
                dntp_conc=self.dntp_conc,
                dna_conc=self.dna_conc,
                temp_c=self.temp_c,
                max_loop=self.max_loop,
            )
            self.assertAlmostEqual(binding_res.tm, wrapper_res.tm, 2)

    def test_correction_methods(self):
        '''Test different correction_methods'''
        self.randArgs()
        for sc_method in ['schildkraut', 'santalucia', 'owczarzy']:
            for tm_method in ['breslauer', 'santalucia']:
                binding_tm = bindings.calc_tm(
                    seq=self.seq1,
                    mv_conc=self.mv_conc,
                    dv_conc=self.dv_conc,
                    dntp_conc=self.dntp_conc,
                    dna_conc=self.dna_conc,
                    tm_method=tm_method,
                    salt_corrections_method=sc_method,
                )
                wrapper_tm = wrappers.calc_tm(
                    seq=self.seq1,
                    mv_conc=self.mv_conc,
                    dv_conc=self.dv_conc,
                    dntp_conc=self.dntp_conc,
                    dna_conc=self.dna_conc,
                    tm_method=tm_method,
                    salt_corrections_method=sc_method,
                )
                self.assertAlmostEqual(binding_tm, wrapper_tm, delta=0.5)

        self.assertRaises(
            ValueError,
            bindings.calc_tm,
            seq=self.seq1,
            mv_conc=self.mv_conc,
            dv_conc=self.dv_conc,
            dntp_conc=self.dntp_conc,
            dna_conc=self.dna_conc,
            tm_method='not_a_tm_method',
        )

    @unittest.skipIf(
        sys.platform == 'win32',
        'Windows does not support resource module',
    )
    def test_memory_leaks(self):
        '''Test for memory leaks'''
        sm = _get_mem_usage()
        run_count = 100
        for x in range(run_count):
            self.randArgs()
            bindings.calc_heterodimer(
                seq1=self.seq1,
                seq2=self.seq2,
                mv_conc=self.mv_conc,
                dv_conc=self.dv_conc,
                dntp_conc=self.dntp_conc,
                dna_conc=self.dna_conc,
                temp_c=self.temp_c,
                max_loop=self.max_loop,
                output_structure=True,
            )
        sleep(0.1)  # Pause for any GC
        em = _get_mem_usage()
        print(
            f'\n\tMemory usage before {run_count} runs of '
            f'calc_heterodimer: {sm}',
        )
        print(
            f'\tMemory usage after {run_count} runs of calc_heterodimer: {em}',
        )
        print(f'\t\t\t\t\tDifference: {em - sm}\t')
        delta_bytes_limit = 500
        if em - sm > delta_bytes_limit:
            raise AssertionError(
                f'Memory usage increase after {run_count} runs of \n\t'
                f'calc_heterodimer > {delta_bytes_limit} bytes -- potential '
                f'\n\t memory leak (mem increase: {em - sm})',
            )


def suite():
    '''Define the test suite'''
    suite = unittest.TestSuite()
    suite.addTest(TestLowLevelBindings())
    return suite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    test_suite = suite()
    runner.run(test_suite)



# Copyright (C) 2023. Ben Pruitt & Nick Conway
# See LICENSE for full GPLv2 license.

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
'''
examples.orthogonalprimers

A simple script to find a complementary oligo set.

NOTE: More advanced desing might use windowed-hamming distance between reference
genome and do more off target checks or look at energy in addition to
melting temperature
'''
import os.path as op
import random
import sys
from typing import (
    Set,
    Tuple,
)

try:
    from primer3 import thermoanalysis
except ModuleNotFoundError:
    PACKAGE_DIR = op.dirname(op.dirname(op.abspath(__file__)))
    sys.path.append(PACKAGE_DIR)
    from primer3 import thermoanalysis


def rev_complement(seq: str) -> str:
    '''Compute the reverse complement of a sequence

    Args:
        seq: 5' to 3' sequence reverse complement

    Returns:
        Reverse complement string
    '''
    lut = {
        'A': 'T',
        'T': 'A',
        'C': 'G',
        'G': 'C',
    }
    return ''.join(lut[x] for x in seq[::-1])


def check_3p_prime_end(seq: str) -> bool:
    '''Check the 3 prime end of an oligo for GC content

    Args:
        seq: sequence to check

    Returns:
        True if GC content at the 3' end is 3 or less else False
    '''
    end3p = seq[-5:]
    return True if end3p.count('G') + end3p.count('C') < 4 else False


def check_gc_content(seq: str, seq_len: int) -> bool:
    '''Check a sequence for total GC content

    Args:
        seq: sequence to check
        seq_len: length sequence string

    Returns:
        True if GC content is less than 50 else False
    '''
    gc_total = seq.count('G') + seq.count('C')
    return True if (gc_total / seq_len) < 0.51 else False


def search_for_30_mers() -> Tuple[int, Set[str]]:
    ''' A simple and inexhaustive search for a set 20-mers that do not
    heterodimerize among themselves or form hairpins or homodimers.

    Returns:
        Tuple of the form::

            <number of candidates found>, <set of oligo strings>
    '''

    # Requirements
    oligo_screen_set_size_limit = 1000  # screen search size
    desired_total_seqs = 200            # break out of loop after 200 are found
    oligo_size = 30                     # we are looking for 30 mers

    # Temperature limits and cutoffs in Celcius
    tm_lim_lo_c = 60
    tm_lim_hi_c = 65
    tm_hairpin_homodimer_cutoff_c = 40
    tm_offtarget_cutoff_c = 40

    # Define analysis parameters and instance
    thermo_params = {
        'mv_conc': 50,      # Monovalent cation concentration in mM
        'dv_conc': 1.5,     # Divalent cation concentration in mM
        'dntp_conc': 0.2,   # dNTP concentration in mM
        'dna_conc': 200,    # DNA concentration in nM
    }
    ta_obj = thermoanalysis.ThermoAnalysis()
    ta_obj.set_thermo_args(**thermo_params)

    # Count the number of qualifying sequences
    found_seqs = 0
    candidate_seq_list = []

    for _ in range(oligo_screen_set_size_limit):
        # Generate a random candidate
        candidate_seq = ''.join([
            random.choice('ATGC') for _ in
            range(oligo_size)
        ])
        if not (
            check_gc_content(candidate_seq, 20) and
            check_3p_prime_end(candidate_seq)
        ):
            continue
        cand_tm_c = ta_obj.calc_tm(candidate_seq)
        cand_hrp_tm = ta_obj.calc_hairpin(candidate_seq).tm
        cand_homo_tm = ta_obj.calc_hairpin(candidate_seq).tm
        if (
            (tm_lim_lo_c < cand_tm_c < tm_lim_hi_c) and
            (cand_hrp_tm > tm_hairpin_homodimer_cutoff_c) and
            (cand_homo_tm > tm_hairpin_homodimer_cutoff_c)
        ):
            candidate_seq_list.append(candidate_seq)
            found_seqs += 1
            if found_seqs > desired_total_seqs:
                break

    # Now check heterodimers and create output greedily
    good_set: Set[str] = set()
    for i, seq1 in enumerate(candidate_seq_list):
        seq1_rev = rev_complement(seq1)

        # NOTE: nothing is done with the bad set but this could be used in a
        # different search pattern
        bad_set: Set[str] = set()

        for _, seq2 in enumerate(candidate_seq_list[i + 1:]):
            tr_obj = ta_obj.calc_heterodimer(seq1, seq2)
            if tr_obj.tm > tm_offtarget_cutoff_c:
                bad_set.add(seq2)
            tr_obj = ta_obj.calc_heterodimer(seq1_rev, seq2)
            if tr_obj.tm > tm_offtarget_cutoff_c:
                bad_set.add(seq2)
        good_set.add(seq1)

    return len(candidate_seq_list), good_set


if __name__ == '__main__':
    number_of_candidates, result_set = search_for_30_mers()
    print(
        f'Found {number_of_candidates} candidates, with total primer '
        f'set size of {len(result_set)}',
    )
    print(f'The {result_set=}')



# Copyright (C) 2023. Ben Pruitt & Nick Conway;
# See LICENSE for full GPLv2 license.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
'''
tests.test_threadsafe
~~~~~~~~~~~~~~~~~~~~~~~~~

Unit tests for the :class:`thermoanalysis.ThermoAnalysis` calls are thread-safe
using the new `thalflex` code

NOTE: It's challenging to get a race condition to compromise thread safety
as the thal() calls are very fast

'''
import random
import threading
import time
import unittest
from typing import (
    Any,
    Dict,
)

from primer3 import thermoanalysis  # type: ignore


def make_random_params() -> Dict[str, Any]:
    '''
    Returns:
        dictionary of random parameters for a call  to
        :meth`ThermoAnalysis.set_thermo_args`

    '''
    return dict(
        mv_conc=random.uniform(1, 200),
        dv_conc=random.uniform(0, 40),
        dntp_conc=random.uniform(0, 20),
        dna_conc=random.uniform(0, 200),
        temp_c=random.randint(10, 70),
        max_loop=random.randint(10, 30),
    )


def make_rand_seq_pair() -> Dict[str, str]:
    '''
    Returns:
        dictionary of 2 unique sequences keyed by seq1 and seq2

    '''
    return dict(
        seq1=''.join([
            random.choice('ATGC') for _ in
            range(random.randint(20, 59))
        ]),
        seq2=''.join([
            random.choice('ATGC') for _ in
            range(random.randint(20, 59))
        ]),
    )


class TestThermoAnalysisInThread(unittest.TestCase):

    def test_calc_heterodimer_thread_safety(self):
        '''Test `calc_heterodimer` (and by extension all thal-dependents) for
        thread safety

        '''
        def het_thread(
                _results_list,
                i,
        ):
            _taf = thermoanalysis.ThermoAnalysis()
            _params = make_random_params()
            _taf.set_thermo_args(
                **_params,
            )
            _seq_pair = make_rand_seq_pair()
            _res = _taf.calc_heterodimer(
                **_seq_pair,
            )
            _results_list[i] = (_res, _params, _seq_pair)

        thread_count = 10
        results_list = [None] * thread_count
        thread_list = []
        t0 = time.time()
        # 1. Run in threads
        for i in range(thread_count):
            t = threading.Thread(
                target=het_thread,
                args=(results_list, i),
            )
            thread_list.append(t)
            t.start()

        for t in thread_list:
            t.join()

        print(f'total: {(time.time()-t0):0.2}')

        taf = thermoanalysis.ThermoAnalysis()
        t0 = time.time()

        # 2. Compare threaded results to sequential results
        for res, params, seq_pair in results_list:
            taf.set_thermo_args(
                **params,
            )
            new_res = taf.calc_heterodimer(
                **seq_pair,
            )
            self.assertEqual(new_res.tm, res.tm)
            self.assertEqual(new_res.dh, res.dh)
        print(f'total: {(time.time()-t0):0.2}')

    def test_primer_design_thread_safety(self):
        '''Test primer design (`run_design`) for thread safety

        '''
        def tdesign_run(_global_args, _results_list, i):
            sequence_template = (
                'GCTTGCATGCCTGCAGGTCGACTCTAGAGGATCCCCCTACATTTTAGCATCAGTGAGTACA'
                'GCATGCTTACTGGAAGAGAGGGTCATGCAACAGATTAGGAGGTAAGTTTGCAAAGGCAGGC'
                'TAAGGAGGAGACGCACTGAATGCCATGGTAAGAACTCTGGACATAAAAATATTGGAAGTTG'
                'TTGAGCAAGTNAAAAAAATGTTTGGAAGTGTTACTTTAGCAATGGCAAGAATGATAGTATG'
                'GAATAGATTGGCAGAATGAAGGCAAAATGATTAGACATATTGCATTAAGGTAAAAAATGAT'
                'AACTGAAGAATTATGTGCCACACTTATTAATAAGAAAGAATATGTGAACCTTGCAGATGTT'
                'TCCCTCTAGTAG'
            ) + ''.join([
                random.choice('ATGC') for _ in
                range(60)
            ])
            quality_list = [
                random.randint(20, 90)
                for i in range(len(sequence_template))
            ]
            seq_args = {
                'SEQUENCE_ID': 'MH1000',
                'SEQUENCE_TEMPLATE': sequence_template,
                'SEQUENCE_QUALITY': quality_list,
                'SEQUENCE_INCLUDED_REGION': (36, 342),
            }
            _taf = thermoanalysis.ThermoAnalysis()
            result = _taf.run_design(
                global_args=_global_args,
                seq_args=seq_args,
                misprime_lib=None,
                mishyb_lib=None,
            )
            _results_list[i] = (result, seq_args, i)

        thread_count = 10
        results_list = [None] * thread_count
        thread_list = []
        t0 = time.time()

        global_args = {
            'PRIMER_OPT_SIZE': 20,
            'PRIMER_PICK_INTERNAL_OLIGO': 1,
            'PRIMER_INTERNAL_MAX_SELF_END': 8,
            'PRIMER_MIN_SIZE': 18,
            'PRIMER_MAX_SIZE': 25,
            'PRIMER_OPT_TM': 60.0,
            'PRIMER_MIN_TM': 57.0,
            'PRIMER_MAX_TM': 63.0,
            'PRIMER_MIN_GC': 20.0,
            'PRIMER_MAX_GC': 80.0,
            'PRIMER_MAX_POLY_X': 100,
            'PRIMER_INTERNAL_MAX_POLY_X': 100,
            'PRIMER_SALT_MONOVALENT': 50.0,
            'PRIMER_DNA_CONC': 50.0,
            'PRIMER_MAX_NS_ACCEPTED': 0,
            'PRIMER_MAX_SELF_ANY': 12,
            'PRIMER_MAX_SELF_END': 8,
            'PRIMER_PAIR_MAX_COMPL_ANY': 12,
            'PRIMER_PAIR_MAX_COMPL_END': 8,
            'PRIMER_PRODUCT_SIZE_RANGE': [
                [75, 100],
                [100, 125],
                [125, 150],
                [150, 175],
                [175, 200],
                [200, 225],
            ],
        }

        # 3. Run in threads
        t0 = time.time()
        for i in range(thread_count):
            t = threading.Thread(
                target=tdesign_run,
                args=(global_args, results_list, i),
            )
            thread_list.append(t)
            t.start()
        print(f'total: {(time.time()-t0):0.2}')

        taf = thermoanalysis.ThermoAnalysis()
        t0 = time.time()
        # 4. Compare threaded results to sequential results
        for res, seq_args, i in results_list:
            new_res = taf.run_design(
                global_args=global_args,
                seq_args=seq_args,
                misprime_lib=None,
                mishyb_lib=None,
            )
            self.assertEqual(new_res, res)
        print(f'total: {(time.time()-t0):0.2}')



# Copyright (C) 2014-2020. Ben Pruitt & Nick Conway; Wyss Institute
# See LICENSE for full GPLv2 license.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
'''
tests._simulatedbindings
~~~~~~~~~~~~~~~~~~~~~~~
'''
import io
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

from primer3 import argdefaults

from . import wrappers

P3_ARGS: Dict[str, Any] = {}

Sequence_T = Union[List[Any], Tuple[Any, ...]]
Str_Bytes_T = Union[str, bytes]


def convert_result(result):
    unwrap = argdefaults.unwrap
    return dict(unwrap(k, v) for k, v in result.items())

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Design bindings ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #


def design_primers(
        seq_args: Dict[str, Any],
        global_args: Optional[Dict[str, Any]] = None,
        reset_args: bool = True,
        misprime_lib: Optional[Dict[Str_Bytes_T, Str_Bytes_T]] = None,
        mishyb_lib: Optional[Dict[Str_Bytes_T, Str_Bytes_T]] = None,
        input_log: Optional[io.BufferedWriter] = None,
        output_log: Optional[io.BufferedWriter] = None,
        err_log: Optional[io.BufferedWriter] = None,
) -> Dict[str, Any]:
    ''' Run the Primer3 design process, with the same interface as the bindings,
    using the wrapped subprocess of primer3_core to do the work.

    If the global args have been previously set (either by a previous
    `design_primers` call or by a `set_globals` call), `design_primers` may be
    called with seqArgs alone (as a means of optimization).

    Args:
        seq_args: Primer3 sequence/design args as per Primer3 docs
        global_args: Primer3 global args as per Primer3 docs
        reset_args:
        misprime_lib: `Sequence name: sequence` dictionary for mispriming
            checks.  NOTE: Unused as of now
        mishyb_lib: `Sequence name: sequence` dictionary for mishybridization
            checks.  NOTE: Unused as of now
        input_log: Optional log input file descriptor
        output_log: Optional log output file descriptor
        err_log: Optional log error file descriptor

    Returns:
        A dictionary of Primer3 results (should be identical to the expected
        BoulderIO output from primer3_main)

    '''
    if reset_args:
        P3_ARGS.clear()

    if global_args:
        P3_ARGS.update(global_args)
    P3_ARGS.update(seq_args)

    result = wrappers.design_primers(
        P3_ARGS,
        input_log=input_log,
        output_log=output_log,
        err_log=err_log,
    )
    return convert_result(result)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# Copyright (C) 2023. Ben Pruitt & Nick Conway
# See LICENSE for full GPLv2 license.

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
'''
examples.basicprimerdesign

A simple script to demonstrate running primer design
'''
import os.path as op
import random
import sys
from typing import (
    Any,
    Dict,
)

try:
    import primer3
except ModuleNotFoundError:
    PACKAGE_DIR = op.dirname(op.dirname(op.abspath(__file__)))
    sys.path.append(PACKAGE_DIR)
    import primer3


def basic_primer_design() -> Dict[str, Any]:
    '''Peform basic primer design given the parameters defined in the function
    below

    Returns:
        Dictionary of the design results
    '''
    sequence_template = (
        'GCTTGCATGCCTGCAGGTCGACTCTAGAGGATCCCCCTACATTTTAGCATCAGTGAGTACAGCAT'
        'GCTTACTGGAAGAGAGGGTCATGCAACAGATTAGGAGGTAAGTTTGCAAAGGCAGGCTAAGGAGG'
        'AGACGCACTGAATGCCATGGTAAGAACTCTGGACATAAAAATATTGGAAGTTGTTGAGCAAGTNA'
        'AAAAAATGTTTGGAAGTGTTACTTTAGCAATGGCAAGAATGATAGTATGGAATAGATTGGCAGAA'
        'TGAAGGCAAAATGATTAGACATATTGCATTAAGGTAAAAAATGATAACTGAAGAATTATGTGCCA'
        'CACTTATTAATAAGAAAGAATATGTGAACCTTGCAGATGTTTCCCTCTAGTAG'
    )
    quality_list = [
        random.randint(20, 90)
        for i in range(len(sequence_template))
    ]
    seq_args = {
        'SEQUENCE_ID': 'MH1000',
        'SEQUENCE_TEMPLATE': sequence_template,
        'SEQUENCE_QUALITY': quality_list,
        'SEQUENCE_INCLUDED_REGION': (36, 342),
    }
    global_args = {
        'PRIMER_OPT_SIZE': 20,
        'PRIMER_PICK_INTERNAL_OLIGO': 1,
        'PRIMER_INTERNAL_MAX_SELF_END': 8,
        'PRIMER_MIN_SIZE': 18,
        'PRIMER_MAX_SIZE': 25,
        'PRIMER_OPT_TM': 60.0,
        'PRIMER_MIN_TM': 57.0,
        'PRIMER_MAX_TM': 63.0,
        'PRIMER_MIN_GC': 20.0,
        'PRIMER_MAX_GC': 80.0,
        'PRIMER_MAX_POLY_X': 100,
        'PRIMER_INTERNAL_MAX_POLY_X': 100,
        'PRIMER_SALT_MONOVALENT': 50.0,
        'PRIMER_DNA_CONC': 50.0,
        'PRIMER_MAX_NS_ACCEPTED': 0,
        'PRIMER_MAX_SELF_ANY': 12,
        'PRIMER_MAX_SELF_END': 8,
        'PRIMER_PAIR_MAX_COMPL_ANY': 12,
        'PRIMER_PAIR_MAX_COMPL_END': 8,
        'PRIMER_PRODUCT_SIZE_RANGE': [
            [75, 100],
            [100, 125],
            [125, 150],
            [150, 175],
            [175, 200],
            [200, 225],
        ],
    }
    design_result_dict = primer3.design_primers(
        seq_args=seq_args,
        global_args=global_args,
    )
    return design_result_dict


if __name__ == '__main__':
    result_dict = basic_primer_design()
    import pprint
    print('Generated the following result')
    pprint.pprint(result_dict)



# Copyright (C) 2023. Ben Pruitt & Nick Conway
# See LICENSE for full GPLv2 license.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
'''
tests.test_argdefaults
~~~~~~~~~~~~~~~~~~~~~~~

Unit tests for the primer3-py argdefaults

'''
import difflib
import os.path as op
import unittest

from primer3 import argdefaults

TEST_DIR = op.dirname(op.realpath(__file__))
TEST_INPUT_DIR = op.join(TEST_DIR, 'input_files')
PACKAGE_DIR = op.dirname(TEST_DIR)


class TestDesignBindings(unittest.TestCase):

    def test_wrap_list_of_quads(self):
        '''
        test wrap_list_of_quads
        '''
        result = argdefaults.wrap_list_of_quads([1, 2, 3, 4])
        self.assertEqual(result, '1,2,3,4')

        result = argdefaults.wrap_list_of_quads([[1, 2, 3, 4]])
        self.assertEqual(result, '1,2,3,4')

        result = argdefaults.wrap_list_of_quads([[1, 2, 3, 4], [5, 6, 7, 8]])
        self.assertEqual(result, '1,2,3,4 ; 5,6,7,8')

        result = argdefaults.wrap_list_of_quads(((1, 2, 3, 4), [5, 6, 7, 8]))
        self.assertEqual(result, '1,2,3,4 ; 5,6,7,8')

        result = argdefaults.wrap_list_of_quads((5, 6, -1, -1))
        print('woof', result)
        self.assertEqual(result, '5,6,,')

        result = argdefaults.wrap_list_of_quads(((1, 2, 3, 4), [5, 6, -1, -1]))
        self.assertEqual(result, '1,2,3,4 ; 5,6,,')

    def test_unwrap(self):
        '''
        Make sure unwrap works correctly
        '''
        _, result = argdefaults.unwrap('', '1,2,3,4 ; 5,6,,')
        self.assertEqual(result, ((1, 2, 3, 4), (5, 6, -1, -1)))

    def test_roundtrip(self):
        '''
        Assert roundtrip conversion to dictionary works
        '''
        test_file_roots = [
            'dv_conc_vs_dntp_conc',
            'long_seq',
            'p3-tmpl-mispriming',
            'primer_all_settingsfiles',
            'primer_check',
            'primer_end_pathology',
            'primer_first_base_index',
            'primer_gc_end',
            'primer_high_gc_load_set',
            'primer_high_tm_load_set',
            'primer_internal',
            'primer_must_overlap_point',
            'primer_must_use_th',
            'primer_num_best',
            'primer_ok_regions',
            'primer_overlap_junction',
            'primer_renewed_tasks',
            'primer_start_codon',
            'primer_task_th',
            'primer_task',
            'primer_thal_args',
            'primer_thal_max_seq_error',
            'primer_tm_lc_masking',
            'test_compl_error',
            'test_left_to_right_of_right',
        ]
        for fn_root in test_file_roots:
            fp = op.join(TEST_INPUT_DIR, f'{fn_root}_input')
            with open(fp, 'r') as fd:
                boulder_str_init = fd.read()

            data_dict_list = argdefaults.parse_multirecord_boulder_io(
                boulder_str_init,
            )
            boulder_str_out = b''
            for data_dict in data_dict_list:
                boulder_str_out += argdefaults.format_boulder_io(
                    data_dict,
                )
            boulder_str_out = boulder_str_out.decode('utf8')
            # Double check that the differences are not just white space
            # and floating point `0` characters
            if boulder_str_init != boulder_str_out:
                for s in difflib.ndiff(boulder_str_init, boulder_str_out):
                    if s[0] == ' ':
                        continue
                    elif s[0] == '-':
                        if s[-1] in (' ', '0'):
                            continue
                    elif s[0] == '+':
                        if s[-1] in (' ', '0'):
                            continue
                    else:
                        # write to file to help with inspection
                        with open('boulder_str_init', 'w') as fd:
                            fd.write(boulder_str_init)
                        with open('boulder_str_out', 'w') as fd:
                            fd.write(boulder_str_out)
                        raise ValueError(f'File difference for {fn_root}')



# Copyright (C) 2014-2020. Ben Pruitt & Nick Conway; Wyss Institute
# See LICENSE for full GPLv2 license.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
'''
tests.test_p3helpers
~~~~~~~~~~~~~~~~~~~~~~~~~

Unit tests for the primer3-py p3helpers.

'''

import unittest

from primer3 import p3helpers  # type: ignore


class TestP3Helpers(unittest.TestCase):

    def test_sanitize_sequence(self):
        # 1. Test upper case
        seq = 'ACGTRYMKSWHDBV'
        out_seq = p3helpers.sanitize_sequence(seq)
        self.assertEqual(out_seq, 'ACGTNNNNNNNNNN')

        # 2. Test lower case
        seq = 'acgtrymkswhdbv'
        out_seq = p3helpers.sanitize_sequence(seq)
        self.assertEqual(out_seq, 'acgtnnnnnnnnnn')

        # 3. Test bytes upper case
        seq = b'ACGTRYMKSWHDBV'
        out_seq = p3helpers.sanitize_sequence_b(seq)
        self.assertEqual(out_seq, b'ACGTNNNNNNNNNN')

        # 4. Test bytes lower case
        seq = b'acgtrymkswhdbv'
        out_seq = p3helpers.sanitize_sequence_b(seq)
        self.assertEqual(out_seq, b'acgtnnnnnnnnnn')

        # 5. Test ValueError for bad base
        self.assertRaises(
            ValueError,
            p3helpers.sanitize_sequence,
            'ZACGT',
        )

    def test_reverse_complement(self):
        # 1. Test upper case no sanitize
        seq = 'ACGTRYMKSWHDBV'
        out_seq = p3helpers.reverse_complement(seq, do_sanitize=False)
        self.assertEqual(out_seq, 'BVHDWSMKRYACGT')

        # 2. Test upper case sanitize
        out_seq = p3helpers.reverse_complement(seq, do_sanitize=True)
        self.assertEqual(out_seq, 'NNNNNNNNNNACGT')

        # 3. Test lower case no sanitize
        seq = 'acgtrymkswhdbv'
        out_seq = p3helpers.reverse_complement(seq, do_sanitize=False)
        self.assertEqual(out_seq, 'bvhdwsmkryacgt')

        # 4. Test lower case sanitize
        out_seq = p3helpers.reverse_complement(seq, do_sanitize=True)
        self.assertEqual(out_seq, 'nnnnnnnnnnacgt')

        # 5. Test bytes upper case no sanitize
        seq = b'ACGTRYMKSWHDBV'
        out_seq = p3helpers.reverse_complement_b(seq, do_sanitize=False)
        self.assertEqual(out_seq, b'BVHDWSMKRYACGT')

        # 5. Test bytes upper case sanitize
        out_seq = p3helpers.reverse_complement_b(seq, do_sanitize=True)
        self.assertEqual(out_seq, b'NNNNNNNNNNACGT')

        # 6. Test bytes lower case no sanitize
        seq = b'acgtrymkswhdbv'
        out_seq = p3helpers.reverse_complement_b(seq, do_sanitize=False)
        self.assertEqual(out_seq, b'bvhdwsmkryacgt')

        # 7. Test bytes lower case sanitize
        out_seq = p3helpers.reverse_complement_b(seq, do_sanitize=True)
        self.assertEqual(out_seq, b'nnnnnnnnnnacgt')

        # 8. Test ValueError for bad base
        self.assertRaises(
            ValueError,
            p3helpers.reverse_complement,
            'ZACGT',
        )


def suite():
    '''Define the test suite'''
    suite = unittest.TestSuite()
    suite.addTest(TestP3Helpers())
    return suite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    test_suite = suite()
    runner.run(test_suite)



# Copyright (C) 2014-2020. Ben Pruitt & Nick Conway; Wyss Institute
# See LICENSE for full GPLv2 license.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
'''
tests.checkdatasets
~~~~~~~~~~~~~~~~~~~
'''
import glob
import json
import os
import os.path
import subprocess
import sys
from typing import (
    Any,
    Dict,
)

import click

from primer3 import __version__ as p3version
from primer3.thermoanalysis import (
    ThermoAnalysis,
    ThermoResult,
)

dirname = os.path.dirname
extra = [dirname(dirname(os.path.abspath(__file__)))]
sys.path = extra + sys.path


def precision(x: float) -> float:
    '''Round the precision of floats to make comparisons exact

    Args:
        x: Float to round

    Returns:
        float rounded to 2 decimal places
    '''
    return round(x, 2)


def todict_tr(x: ThermoResult) -> Dict[str, Any]:
    '''Convert ``ThermoResult`` instance to a dictionary.
    Round float values for readability

    Args:
        x: ``ThermoResult`` instance

    Returns:
        Dictionary form of a ``ThermoResult``
    '''
    return {
        'structure_found': x.structure_found,
        'tm': precision(x.tm),
        'dg': precision(x.dg),
        'dh': precision(x.dh),
        'ds': precision(x.ds),
    }


def todict_ta(x: ThermoAnalysis) -> Dict[str, Any]:
    '''Convert ``ThermoAnalysis`` instance to a dictionary.

    Args:
        x: ``ThermoAnalysis`` instance

    Returns:
        Dictionary form of a ``ThermoAnalysis``
    '''
    return {
        'mv_conc': x.mv_conc,
        'dv_conc': x.dv_conc,
        'dntp_conc': x.dntp_conc,
        'dna_conc': x.dna_conc,
        'temp_c': x.temp,
        'max_loop': x.max_loop,
        # 'temp_only': x.temp_only,
        # 'debug': x.thalargs.debug,
        'max_nn_length': x.max_nn_length,
        'tm_method': x.tm_method,
        'salt_correction_method': x.salt_correction_method,
    }


def git_hash() -> str:
    '''
    Returns:
        Git hash of repo or NA if not available
    '''
    try:
        with open(os.devnull, 'wb') as devnull:
            hash_str = subprocess.check_output(
                ['git', 'rev-parse', '--short', 'HEAD'],
                stderr=devnull,
            )
        return hash_str.strip().decode('utf8')
    except BaseException:
        return 'NA'
# end def


def create_data_set() -> Dict[str, Any]:
    '''Create data sets for testing

    Returns:
        Dictionary of test dataset dictionaries
    '''
    seq1 = 'GGGGCCCCCCAAATTTTTT'
    seq2 = 'AAAAAATTTGGCCCCAAA'

    ta = ThermoAnalysis()
    tests = [
        ['calc_heterodimer', seq1, seq2],
        ['calc_hairpin', seq1],
        ['calc_homodimer', seq1],
        ['calc_tm', seq1],
    ]
    cases = []
    for test in tests:
        f = getattr(ta, test[0])
        res = f(*test[1:])
        if isinstance(res, ThermoResult):
            cases.append([test, todict_tr(res)])
        else:
            cases.append([test, res])
    out = {
        'thermoanalysis': todict_ta(ta),
        'cases': cases,
    }
    return out


def dump_data_set(x: Dict[str, Any]) -> str:
    '''
    Dump data set to a JSON file with the data time and git hash

    Args:
        x: Data set dictionary to dump

    Returns:
        Filename formatted in form::

        `primer3_precision_<date>_<githash>_<OS-platform>_<Python-version>.json`

    '''
    import time
    date_str = time.strftime('%Y-%m-%d-%H-%M-%S', time.gmtime())
    git_hash_str = git_hash()
    py_version = '%s.%s.%s' % sys.version_info[0:3]
    platform = sys.platform
    filename = 'primer3_precision_%s_%s_%s_%s.json' % (
        date_str, git_hash_str, platform, py_version,
    )
    with open(filename, 'w') as fd:
        print('dumping')
        json.dump(x, fd)
    return filename


def read_data_set() -> Dict[str, Any]:
    '''
    Read data set JSON file

    Returns:
        Dictionary of a the first JSON file found in the local directory
    '''
    files = glob.glob('*.json')
    if len(files) > 0:
        the_file = files[0]
        file_name = os.path.splitext(the_file)[0]
        git_hash, platform, py_version = file_name.split('_')[3:6]
        print('\t- Checking reference file: %s' % the_file)
        print(
            '\t\tgit hash: %s\n\t\tPlatform: %s\n\t\tPython: %s' %
            (git_hash, platform, py_version),
        )
        with open(the_file, 'r') as fd:
            out = json.load(fd)
        return out
    else:
        raise OSError('No json file found')


def compare_sets(
        reference: Dict[str, Any],
        test_set: Dict[str, Any],
) -> bool:
    '''
    Compare thermoanalysis datasets

    Args:
        reference: Reference thermoanalysis dataset
        test_set: Test thermoanalysis data set

    Returns:
        True if the same, False otherwise
    '''
    for k1, v in reference.items():
        if k1 not in test_set:
            return False
        test_val = test_set[k1]
        if k1 == 'thermoanalysis':
            for k2, item in v.items():
                if item != test_val[k2]:
                    return False
        elif k1 == 'cases':
            for item_ref, item_test in zip(v, test_val):
                ref_args, ref_result = item_ref
                test_args, test_result = item_test
                if ref_args != test_args:
                    print(ref_args, test_args)
                    return False
                if ref_result != test_result:
                    print(ref_result, test_result)
                    return False

    return True


@click.command()
@click.option('--generate', default=False, help='Generate a set')
def runner(generate):
    create_dict = create_data_set()
    if generate:
        filename = dump_data_set(create_dict)
        print('> Dumped data set: %s' % (filename))
    else:
        print('> Comparing reference and current revisions')
        print('\t- Current git hash: %s' % (git_hash()))
        print('\t- Current primer3 version: %s' % (p3version))
        read_dict = read_data_set()
        print('\t- Is the reference set equal to the current revisions set?')
        are_equal = compare_sets(read_dict, create_dict)
        print('\t\t%s' % (str(are_equal)))


if __name__ == '__main__':
    runner()



# Copyright (C) 2014-2020. Ben Pruitt & Nick Conway; Wyss Institute
# See LICENSE for full GPLv2 license.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
'''
tests.wrappers
~~~~~~~~~~~~~~~~

Simple subprocess wrappers for Primer3 executables. These functions closely
mirror the functions found in bindings.py, but are much slower and should
only be used for testing / comparison purposes.

'''

from __future__ import print_function

import io
import os
import re
import subprocess
import sys
from collections import namedtuple
from os.path import join as pjoin
from typing import (
    Any,
    Dict,
    Optional,
    Tuple,
    Union,
)

from primer3.argdefaults import (
    Primer3PyArguments,
    format_boulder_io,
    parse_boulder_io,
)

DEFAULT_P3_ARGS = Primer3PyArguments()

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ PRIMER3 WRAPPERS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #

# TODO: remove after update to primer3 >= 2.5.0
LOCAL_DIR = os.path.dirname(os.path.realpath(__file__))
PACKAGE_DIR = os.path.dirname(LOCAL_DIR)
LIBPRIMER3_PATH = pjoin(PACKAGE_DIR, 'primer3', 'src', 'libprimer3')
THERMO_PATH = pjoin(
    LIBPRIMER3_PATH,
    'primer3_config',
    '',  # Add trailing slash (OS-ind) req'd by primer3 lib
)

BINARY_EXT = '.exe' if sys.platform == 'win32' else ''

DEV_NULL = open(os.devnull, 'wb')

_tm_methods = {
    'breslauer': 0,
    'santalucia': 1,
}

_salt_corrections_methods = {
    'schildkraut': 0,
    'santalucia': 1,
    'owczarzy': 2,
}


def calc_tm(
        seq: str,
        mv_conc: Union[float, int] = DEFAULT_P3_ARGS.mv_conc,
        dv_conc: Union[float, int] = DEFAULT_P3_ARGS.dv_conc,
        dntp_conc: Union[float, int] = DEFAULT_P3_ARGS.dntp_conc,
        dna_conc: Union[float, int] = DEFAULT_P3_ARGS.dna_conc,
        dmso_conc: float = DEFAULT_P3_ARGS.dmso_conc,
        dmso_fact: float = DEFAULT_P3_ARGS.dmso_fact,
        formamide_conc: float = DEFAULT_P3_ARGS.formamide_conc,
        annealing_temp_c: float = DEFAULT_P3_ARGS.annealing_temp_c,
        max_nn_length: int = DEFAULT_P3_ARGS.max_nn_length,
        tm_method: str = DEFAULT_P3_ARGS.tm_method,
        salt_corrections_method: str = DEFAULT_P3_ARGS.salt_corrections_method,
) -> float:
    ''' Return the tm of `seq` as a float.

    Args:
        seq: DNA sequence
        mv_conc: Monovalent cation conc. (mM)
        dv_conc: Divalent cation conc. (mM)
        dntp_conc: dNTP conc. (mM)
        dna_conc: DNA conc. (nM)
        dmso_conc: Concentration of DMSO (%)
        dmso_fact: DMSO correction factor, default 0.6
        formamide_conc: Concentration of formamide (mol/l)
        annealing_temp_c: Actual annealing temperature of the PCR reaction
            in (C)
        max_nn_length: Maximum length for nearest-neighbor calcs
        tm_method: Tm calculation method (breslauer or santalucia)
        salt_corrections_method: Salt correction method (schildkraut, owczarzy,
            santalucia)

    Returns:
        The melting temperature in degrees Celsius (float).
    '''
    tm_meth = _tm_methods.get(tm_method)
    if tm_meth is None:
        raise ValueError(
            f'{tm_method} is not a valid tm calculation method',
        )
    salt_meth = _salt_corrections_methods.get(salt_corrections_method)
    if salt_meth is None:
        raise ValueError(
            f'{salt_corrections_method} is not a valid salt correction method',
        )
    # For whatever reason mv_conc and dna_conc have to be ints
    args = [
        pjoin(LIBPRIMER3_PATH, f'oligotm{BINARY_EXT}'),
        '-mv', str(mv_conc),
        '-dv', str(dv_conc),
        '-n', str(dntp_conc),
        '-d', str(dna_conc),
        '-tp', str(tm_meth),
        '-sc', str(salt_meth),
        '-dm', str(dmso_conc),
        '-df', str(dmso_fact),
        '-fo', str(formamide_conc),
        seq,
    ]
    tm = subprocess.check_output(
        args, stderr=DEV_NULL,
        env=os.environ,
    )
    return float(tm)


calcTm = calc_tm


_NTTHAL_RE = re.compile(
    r'dS\s+=\s+(\S+)\s+'
    r'dH\s+=\s+(\S+)\s+'
    r'dG\s+=\s+(\S+)\s+'
    r't\s+=\s+(\S+)'.encode('utf8'),
)

THERMORESULT = namedtuple(  # type: ignore
    'thermoresult', [
        'result',           # True if a structure is present
        'ds',               # Entropy (cal/(K*mol))
        'dh',               # Enthalpy (kcal/mol)
        'dg',               # Gibbs free energy
        'tm',               # Melting temperature (deg. Celsius)
        'ascii_structure',   # ASCII representation of structure
    ],
)

NULLTHERMORESULT = THERMORESULT(False, 0, 0, 0, 0, '')


def _parse_ntthal(ntthal_output: bytes) -> THERMORESULT:
    ''' Helper method that uses regex to parse ntthal output. '''
    parsed_vals = re.search(_NTTHAL_RE, ntthal_output)
    if parsed_vals:
        ascii_structure = (
            ntthal_output[ntthal_output.index(b'\n') + 1:].decode('utf8')
        )
        res = THERMORESULT(
            True,                           # Structure found
            float(parsed_vals.group(1)),    # dS
            float(parsed_vals.group(2)),    # dH
            float(parsed_vals.group(3)),    # dG
            float(parsed_vals.group(4)),    # tm
            ascii_structure,
        )
    else:
        res = NULLTHERMORESULT
    return res


def calc_thermo(
        seq1: str,
        seq2: str,
        calc_type: str = DEFAULT_P3_ARGS.calc_type_wrapper,
        mv_conc: Union[float, int] = DEFAULT_P3_ARGS.mv_conc,
        dv_conc: Union[float, int] = DEFAULT_P3_ARGS.dv_conc,
        dntp_conc: Union[float, int] = DEFAULT_P3_ARGS.dntp_conc,
        dna_conc: Union[float, int] = DEFAULT_P3_ARGS.dna_conc,
        temp_c: Union[float, int] = DEFAULT_P3_ARGS.temp_c,
        max_loop: int = DEFAULT_P3_ARGS.max_loop,
        temp_only: Union[bool, int] = DEFAULT_P3_ARGS.temp_only,
) -> THERMORESULT:
    '''Main subprocess wrapper for calls to the ntthal executable.

    Args:
        seq1: DNA sequence to analyze for 3' end  hybridization against the
            target sequence
        seq2: Target DNA sequence to analyze for seq1 3' end hybridization
        calc_type: alignment type, END1, END2, ANY and HAIRPIN, by default ANY
            (when duplex)
        mv_conc: Monovalent cation conc. (mM)
        dv_conc: Divalent cation conc. (mM)
        dntp_conc: dNTP conc. (mM)
        dna_conc: DNA conc. (nM)
        temp_c: Simulation temperature for dG (Celsius)
        max_loop: Maximum size of loops in the structure
        temp_only: causes the alignment NOT to be displayed on stderr,
            _only_ Tm is printed

    Returns:
        a named tuple with tm, ds, dh, and dg values or None if no
        structure / complex could be computed.
    '''
    args = [
        pjoin(LIBPRIMER3_PATH, f'ntthal{BINARY_EXT}'),
        '-a', str(calc_type),
        '-mv', str(mv_conc),
        '-dv', str(dv_conc),
        '-n', str(dntp_conc),
        '-d', str(dna_conc),
        '-t', str(temp_c),
        '-maxloop', str(max_loop),
        '-path', THERMO_PATH,
        '-s1', seq1,
        '-s2', seq2,
    ]
    if temp_only:
        args += ['-r']
    out = subprocess.check_output(
        args, stderr=DEV_NULL,
        env=os.environ,
    )
    return _parse_ntthal(out)


def calc_hairpin(
        seq: str,
        mv_conc: Union[float, int] = DEFAULT_P3_ARGS.mv_conc,
        dv_conc: Union[float, int] = DEFAULT_P3_ARGS.dv_conc,
        dntp_conc: Union[float, int] = DEFAULT_P3_ARGS.dntp_conc,
        dna_conc: Union[float, int] = DEFAULT_P3_ARGS.dna_conc,
        temp_c: Union[float, int] = DEFAULT_P3_ARGS.temp_c,
        max_loop: int = DEFAULT_P3_ARGS.max_loop,
        temp_only: Union[bool, int] = DEFAULT_P3_ARGS.temp_only,
) -> THERMORESULT:
    ''' Return a namedtuple of the dS, dH, dG, and Tm of any hairpin struct
    present.

    Args:
        seq: DNA sequence to analyze for hairpin formation
        mv_conc: Monovalent cation conc. (mM)
        dv_conc: Divalent cation conc. (mM)
        dntp_conc: dNTP conc. (mM)
        dna_conc: DNA conc. (nM)
        temp_c: Simulation temperature for dG (Celsius)
        max_loop(int, optional): Maximum size of loops in the structure
        temp_only: print only temp to stderr

    Returns:
       ``THERMORESULT`` tuple
    '''
    return calc_thermo(
        seq,
        seq,
        'HAIRPIN',
        mv_conc,
        dv_conc,
        dntp_conc,
        dna_conc,
        temp_c,
        max_loop,
        temp_only,
    )


calcHairpin = calc_hairpin


def calc_heterodimer(
        seq1: str,
        seq2: str,
        mv_conc: Union[float, int] = DEFAULT_P3_ARGS.mv_conc,
        dv_conc: Union[float, int] = DEFAULT_P3_ARGS.dv_conc,
        dntp_conc: Union[float, int] = DEFAULT_P3_ARGS.dntp_conc,
        dna_conc: Union[float, int] = DEFAULT_P3_ARGS.dna_conc,
        temp_c: Union[float, int] = DEFAULT_P3_ARGS.temp_c,
        max_loop: int = DEFAULT_P3_ARGS.max_loop,
        temp_only: Union[bool, int] = DEFAULT_P3_ARGS.temp_only,
) -> THERMORESULT:
    '''Returns a tuple of the dS, dH, dG, and Tm of any predicted heterodimer.

    Args:
        seq1: DNA sequence to analyze for heterodimer formation
        seq2: DNA sequence to analyze for heterodimer formation
        mv_conc: Monovalent cation conc. (mM)
        dv_conc: Divalent cation conc. (mM)
        dntp_conc: dNTP conc. (mM)
        dna_conc: DNA conc. (nM)
        temp_c: Simulation temperature for dG (Celsius)
        max_loop(int, optional): Maximum size of loops in the structure
        temp_only:
        temp_only: print only temp to stderr

    Returns:
       ``THERMORESULT`` tuple
    '''
    return calc_thermo(
        seq1,
        seq2,
        'ANY',
        mv_conc,
        dv_conc,
        dntp_conc,
        dna_conc,
        temp_c,
        max_loop,
        temp_only,
    )


calcHeterodimer = calc_heterodimer


def calc_homodimer(
        seq: str,
        mv_conc: Union[float, int] = DEFAULT_P3_ARGS.mv_conc,
        dv_conc: Union[float, int] = DEFAULT_P3_ARGS.dv_conc,
        dntp_conc: Union[float, int] = DEFAULT_P3_ARGS.dntp_conc,
        dna_conc: Union[float, int] = DEFAULT_P3_ARGS.dna_conc,
        temp_c: Union[float, int] = DEFAULT_P3_ARGS.temp_c,
        max_loop: int = DEFAULT_P3_ARGS.max_loop,
        temp_only: Union[bool, int] = DEFAULT_P3_ARGS.temp_only,
) -> THERMORESULT:
    ''' Return a tuple of the dS, dH, dG, and Tm of any predicted homodimer.

    Args:
        seq: DNA sequence to analyze for homodimer formation
        mv_conc: Monovalent cation conc. (mM)
        dv_conc: Divalent cation conc. (mM)
        dntp_conc: dNTP conc. (mM)
        dna_conc: DNA conc. (nM)
        temp_c: Simulation temperature for dG (Celsius)
        max_loop(int, optional): Maximum size of loops in the structure
        temp_only:
        temp_only: print only temp to stderr

    Returns:
       ``THERMORESULT`` tuple
    '''
    return calc_thermo(
        seq,
        seq,
        'ANY',
        mv_conc,
        dv_conc,
        dntp_conc,
        dna_conc,
        temp_c,
        max_loop,
        temp_only,
    )


calcHomodimer = calc_homodimer


def calc_end_stability(
        seq1: str,
        seq2: str,
        mv_conc: Union[float, int] = DEFAULT_P3_ARGS.mv_conc,
        dv_conc: Union[float, int] = DEFAULT_P3_ARGS.dv_conc,
        dntp_conc: Union[float, int] = DEFAULT_P3_ARGS.dntp_conc,
        dna_conc: Union[float, int] = DEFAULT_P3_ARGS.dna_conc,
        temp_c: Union[float, int] = DEFAULT_P3_ARGS.temp_c,
        max_loop: int = DEFAULT_P3_ARGS.max_loop,
        temp_only: Union[bool, int] = DEFAULT_P3_ARGS.temp_only,
) -> THERMORESULT:
    ''' Return a tuple of the dS, dH, dG, and Tm of any predicted heterodimer
    end stability

    Args:
        seq1: DNA sequence to analyze for end stability
        seq2: DNA sequence to analyze for end stability
        mv_conc: Monovalent cation conc. (mM)
        dv_conc: Divalent cation conc. (mM)
        dntp_conc: dNTP conc. (mM)
        dna_conc: DNA conc. (nM)
        temp_c: Simulation temperature for dG (Celsius)
        max_loop(int, optional): Maximum size of loops in the structure
        temp_only:
        temp_only: print only temp to stderr

    Returns:
       ``THERMORESULT`` tuple
    '''
    return calc_thermo(
        seq1,
        seq2,
        'END1',
        mv_conc,
        dv_conc,
        dntp_conc,
        dna_conc,
        temp_c,
        max_loop,
        temp_only,
    )


calcEndStability = calc_end_stability


def assess_oligo(seq: str) -> Tuple[THERMORESULT, THERMORESULT]:
    '''
    Return the thermodynamic characteristics of hairpin/homodimer structures.

    Args:
        seq: DNA sequence to analyze

    Returns:
        A tuple of namedtuples (hairpin data, homodimer data) in which each
        individual tuple is structured (dS, dH, dG, Tm).

    '''
    hairpin_out = calc_hairpin(seq)
    homodimer_out = calc_homodimer(seq)
    return (hairpin_out, homodimer_out)


def design_primers(
        p3_args: Dict[str, Any],
        input_log: Optional[io.BufferedWriter] = None,
        output_log: Optional[io.BufferedWriter] = None,
        err_log: Optional[io.BufferedWriter] = None,
) -> Dict[str, Any]:
    ''' Return the raw primer3_core output for the provided primer3 args.

    Args:
        p3_args: Dictionarty of arguments for the primer3
        input_log: Optional log input file descriptor
        output_log: Optional log output file descriptor
        err_log: Optional log error file descriptor

    Returns:
        an ordered dict of the boulderIO-format primer3 output file
    '''
    sp = subprocess.Popen(
        [pjoin(LIBPRIMER3_PATH, f'primer3_core{BINARY_EXT}')],
        stdout=subprocess.PIPE,
        stdin=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    p3_args.setdefault('PRIMER_THERMODYNAMIC_PARAMETERS_PATH', THERMO_PATH)
    in_str = format_boulder_io(p3_args)
    if input_log:
        input_log.write(in_str)
        input_log.flush()
    out_str, err_str = sp.communicate(input=in_str)
    if output_log:
        output_log.write(out_str)
        output_log.flush()
    if err_log and err_str is not None:
        err_log.write(err_str)
        err_log.flush()
    return parse_boulder_io(out_str)


designPrimers = design_primers



# cython: language_level=3
# Copyright (C) 2023. Ben Pruitt & Nick Conway;
# See LICENSE for full GPLv2 license.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
'''
primer3.p3helpers
~~~~~~~~~~~~~~~~~~~~~~

Contains Cython functions and classes that aid primer design

'''
from cpython.mem cimport (
    PyMem_Free,
    PyMem_Malloc,
)
from libc.string cimport memcpy


cdef extern from "p3helpers.h":
    const char COMP_BASE_LUT[128]
    const char SANITIZE_LUT[128]


cdef:
    char INVALID_BASE = 63 # '?' character


cdef inline void copy_c_char_buffer(
    char*   in_buf,
    char*   out_buf,
    int     str_length,
):
    '''Copy string to buffer.  Unsafe

    Args:
        in_buf: input buffer
        out_buf: output buffer. Assume buffer size is equal to or greater than
            in_buf size.
        str_length: known number of bytes to copy
    '''
    memcpy(out_buf, in_buf, <size_t> str_length + 1)


cdef int ss_rev_comp_seq(char* seq, Py_ssize_t length) nogil:
    '''Reverse complement sequence C string

    Args:
        seq: sequence to sanitize
        length: length of sequence

    Returns:
        0 on success, 1 on error (Invalid base)
    '''
    cdef:
        char        temp
        char*       end_ptr = NULL

    end_ptr = seq + (length - 1)

    while end_ptr > seq:
        temp = end_ptr[0]
        end_ptr[0] = COMP_BASE_LUT[<unsigned char> seq[0]]
        if end_ptr[0] == INVALID_BASE:
            return 1
        end_ptr -= 1
        seq[0] = COMP_BASE_LUT[<unsigned char> temp]
        if seq[0] == INVALID_BASE:
            return 1
        seq += 1


    if length % 2:
        seq[0] = COMP_BASE_LUT[<unsigned char> seq[0]]

    return 0


cdef int ss_sanitize_seq(
    char* seq,
    Py_ssize_t length,
) nogil:
    '''Sanitize sequence C string IUPAC non-{A,C,G,T,a,c,g,t} bases with {N, n}

    Args:
        seq: sequence to sanitize
        length: length of sequence

    Returns:
        0 on success, 1 on error (Invalid base)
    '''
    cdef:
        char*       end_ptr = seq + length

    while seq < end_ptr:
        seq[0] = SANITIZE_LUT[<unsigned char> seq[0]]
        if seq[0] == INVALID_BASE:
            return 1
        seq += 1
    return 0


def reverse_complement(
        seq: str,
        bint do_sanitize = False,
) -> str:
    '''Compute reverse complement of the python string sequence

    Args:
        seq: sequence string
        do_sanitize: If True, convert non-IUPAC characters to N's

    Returns:
        Reverse complement of sequence string

    Raises:
        ValueError: Invalid base in sequence
    '''
    cdef:
        char* seq_c = NULL
        Py_ssize_t seq_len = len(seq)
        int check = 0

    seq_b = seq.encode('utf8')
    # convert to C str
    seq_c = seq_b

    check = ss_rev_comp_seq(seq_c, seq_len)
    if check:
        raise ValueError(f'Invalid base in sequence {seq}')
    if do_sanitize:
        check = ss_sanitize_seq(seq_c, seq_len)
        if check:
            raise ValueError(f'Invalid base in sequence {seq}')
    seq_b = seq_c

    return seq_b.decode('utf8')


def reverse_complement_b(
        seq: bytes,
        bint do_sanitize = False,
) -> bytes:
    '''Compute reverse complement of the bytes sequence

    Args:
        seq: sequence in bytes
        do_sanitize: If True, convert non-IUPAC characters to N's

    Returns:
        Reverse complement of sequence in bytes

    Raises:
        ValueError: Invalid base in sequence
        OSError: ``malloc`` failure
    '''
    cdef:
        char* seq_c = NULL
        Py_ssize_t seq_len = len(seq)
        char* seq_operate_c = NULL
        bytes seq_out
        int check = 0

    seq_operate_c = <char *> PyMem_Malloc(seq_len * sizeof(char))
    if seq_operate_c == NULL:
        raise OSError('malloc failure')

    # convert to C str
    seq_c = seq

    # MUST COPY as python does not mutate bytes
    copy_c_char_buffer(seq_c, seq_operate_c, seq_len + 1)

    check = ss_rev_comp_seq(seq_operate_c, seq_len)
    if check:
        raise ValueError(f'Invalid base in sequence {seq}')
    if do_sanitize:
        check = ss_sanitize_seq(seq_operate_c, seq_len)
        if check:
            PyMem_Free(seq_operate_c)
            seq_operate_c = NULL
            raise ValueError(f'Invalid base in sequence {seq}')
    seq_out = seq_operate_c
    PyMem_Free(seq_operate_c)
    seq_operate_c = NULL
    return seq_out


def sanitize_sequence(seq: str) -> str:
    '''Sanitize sequence with non-standard bases with `N`s
    IUPAC {R,Y,M,K,S,W,H,D,B,V,r,y,m,k,s,w,h,d,b,v}

    IUPAC non-{A,C,G,T,a,c,g,t} bases become {N, n}

    NOTE: Consider keeping a copy of original ``seq`` argument for record
    keeping

    Args:
        seq: sequence to sanitize nonstandard with `N`s or `n`s

    Returns:
        sanitized version of ``seq``

    Raises:
        ValueError: Invalid base in sequence
    '''
    cdef:
        char* seq_c = NULL
        Py_ssize_t seq_len = len(seq)
        int check = 0

    seq_b = seq.encode('utf8')
    # convert to C str
    seq_c = seq_b

    check = ss_sanitize_seq(seq_c, seq_len)
    if check:
        raise ValueError(f'Invalid base in equence {seq}')
    seq_b = seq_c

    return seq_b.decode('utf8')


def sanitize_sequence_b(
        seq: bytes,
) -> bytes:
    '''Sanitize bytes sequence with non-standard bases with `N`s
    IUPAC {R,Y,M,K,S,W,H,D,B,V,r,y,m,k,s,w,h,d,b,v}

    IUPAC non-{A,C,G,T,a,c,g,t} bases become {N, n}

    NOTE: Consider keeping a copy of original ``seq`` argument for record
    keeping

    Args:
        seq: sequence to sanitize nonstandard with `N`s or `n`s

    Returns:
        sanitized version of ``seq``

    Raises:
        ValueError: Invalid base in sequence
        OSError: ``malloc`` failure
    '''
    cdef:
        char* seq_c = NULL
        Py_ssize_t seq_len = len(seq)
        char* seq_operate_c = NULL
        bytes seq_out
        int check = 0

    seq_operate_c = <char *> PyMem_Malloc(seq_len * sizeof(char))
    if seq_operate_c == NULL:
        raise OSError('malloc failure')

    # convert to C str
    seq_c = seq

    # MUST COPY as python does not mutate bytes
    copy_c_char_buffer(seq_c, seq_operate_c, seq_len + 1)

    check = ss_sanitize_seq(seq_operate_c, seq_len)
    if check:
        PyMem_Free(seq_operate_c)
        seq_operate_c = NULL
        raise ValueError(f'Invalid base in equence {seq}')
    seq_out = seq_operate_c
    PyMem_Free(seq_operate_c)
    seq_operate_c = NULL
    return seq_out



# Copyright (C) 2014-2020. Ben Pruitt & Nick Conway; Wyss Institute
# See LICENSE for full GPLv2 license.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
'''
tests.test_primerdesign
~~~~~~~~~~~~~~~~~~~~~~~

Unit tests for the primer3-py primer design bindings.

'''

from __future__ import print_function

import os
import random
import sys
import unittest
from time import sleep
from typing import (
    Any,
    Dict,
    List,
    Tuple,
)

try:
    import resource
except (ImportError, ModuleNotFoundError):  # For Windows compatibility
    resource = None  # type: ignore

from primer3 import (
    argdefaults,
    bindings,
)

from . import _simulatedbindings as simulatedbindings

LOCAL_DIR = os.path.dirname(os.path.realpath(__file__))


def _get_mem_usage():
    ''' Get current process memory usage in bytes '''
    return resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024


class TestDesignBindings(unittest.TestCase):

    def _compare_results(
        self,
        binding_res: Dict[str, Any],
        simulated_binding_res: Dict[str, Any],
        verbose: bool = False,
    ) -> str:
        '''Compare results between binding and simulated binding calls to
        ``design_primers``

        Args:
            binding_res: A dictionary of Primer3 results (should be identical
                to the expected BoulderIO output from primer3_main) from the
                bindings call
            simulated_binding_res: A dictionary of Primer3 results (should be
                identical to the expected BoulderIO output from primer3_main)
                from the simulated bindings call
            verbose: if True, print information

        Returns:
            String of disagreements between the binding and simulated results
        '''
        keys_in_sim = set(simulated_binding_res)
        keys_in_binding = set(binding_res)

        if keys_in_sim - keys_in_binding:
            if verbose:
                print(
                    '\n\n\nIn wrapper simulation result but missing'
                    ' from binding:',
                )
                fmt = '{:<30} {:<50}'
                print(fmt.format('Output Key', 'SimBinding Result'))
                print('-' * 80)
                for k in sorted(keys_in_sim - keys_in_binding):
                    print(fmt.format(k, repr(simulated_binding_res[k])))

        if keys_in_binding - keys_in_sim:
            if verbose:
                print(
                    '\n\n\nIn binding result but missing from wrapper '
                    'simulation:',
                )
                fmt = '{:<30} {:<50}'
                print(fmt.format('Output Key', 'Binding Result'))
                print('-' * 80)
                for k in sorted(keys_in_binding - keys_in_sim):
                    print(fmt.format(k, repr(binding_res[k])))

        allowable_relative_difference = 0.05
        discrepencies: List[str] = [
            k for k in keys_in_binding & keys_in_sim
            if simulated_binding_res[k] != binding_res[k]
        ]
        disagreements_list: List[str] = []
        for ds in discrepencies:
            if (
                isinstance(binding_res[ds], (float, int)) and
                binding_res[ds] != 0
            ):
                percent_diff = abs(
                    (binding_res[ds] - simulated_binding_res[ds]) /
                    binding_res[ds],
                )
                if percent_diff > allowable_relative_difference:
                    if simulated_binding_res[ds] == 0.0 and binding_res[ds] < 0:
                        pass
                    else:
                        disagreements_list.append(ds)

        if len(disagreements_list):
            fmt = '{:<30} {:<25} {:<25}'
            disagreements_str = '\n'.join([
                fmt.format(
                    k,
                    repr(simulated_binding_res[k]),
                    repr(binding_res[k]),
                ) for k in sorted(disagreements_list)
            ])
            if verbose:
                print('\n\n\nResults disagree:')
                print(
                    fmt.format(
                        'Output Key',
                        'SimBinding Result',
                        'Binding Result',
                    ),
                )
                print('-' * 80)
            return disagreements_str
        else:
            if verbose:
                print(
                    '\n\n\nAll the results in common '
                    f'({len(keys_in_binding & keys_in_sim)}) agree to within '
                    f'{allowable_relative_difference:.2%}',
                )
            return ''

    def _convert_boulder_input(
            self,
            boulder_str: str,
    ) -> List[Tuple[Dict, Dict, Dict]]:
        ''' Convert a boulder IO-style input dictionary into bindings /
        simulated-bindings-friendly dictionaries.

        Args:
            boulder_str: Boulder formatted string of multiple records

        Returns:
            List of a tuple of dictionaries of the form::

            [
                (global_args, seq_args, p3_args),
                ...
            ]
        '''
        boulder_dicts = argdefaults.parse_multirecord_boulder_io(boulder_str)
        input_dicts_list = []
        for bd in boulder_dicts:
            converted_input = [(k, v) for k, v in bd.items()]
            global_args = dict(
                filter(
                    lambda arg: 'PRIMER_' == arg[0][:7],
                    converted_input,
                ),
            )
            seq_args = dict(
                filter(
                    lambda arg: 'SEQUENCE_' == arg[0][:9],
                    converted_input,
                ),
            )
            p3_args = dict(
                filter(
                    lambda arg: 'P3_' == arg[0][:3],
                    converted_input,
                ),
            )
            input_dicts_list.append((global_args, seq_args, p3_args))
        return input_dicts_list

    def test_compare_sim(self):
        sequence_template = (
            'GCTTGCATGCCTGCAGGTCGACTCTAGAGGATCCCCCTACATTTTAGCATCAGTGAGTACAGCAT'
            'GCTTACTGGAAGAGAGGGTCATGCAACAGATTAGGAGGTAAGTTTGCAAAGGCAGGCTAAGGAGG'
            'AGACGCACTGAATGCCATGGTAAGAACTCTGGACATAAAAATATTGGAAGTTGTTGAGCAAGTNA'
            'AAAAAATGTTTGGAAGTGTTACTTTAGCAATGGCAAGAATGATAGTATGGAATAGATTGGCAGAA'
            'TGAAGGCAAAATGATTAGACATATTGCATTAAGGTAAAAAATGATAACTGAAGAATTATGTGCCA'
            'CACTTATTAATAAGAAAGAATATGTGAACCTTGCAGATGTTTCCCTCTAGTAG'
        )
        quality_list = [
            random.randint(20, 90)
            for i in range(len(sequence_template))
        ]
        seq_args = {
            'SEQUENCE_ID': 'MH1000',
            'SEQUENCE_TEMPLATE': sequence_template,
            'SEQUENCE_QUALITY': quality_list,
            'SEQUENCE_INCLUDED_REGION': (36, 342),
        }
        global_args = {
            'PRIMER_OPT_SIZE': 20,
            'PRIMER_PICK_INTERNAL_OLIGO': 1,
            'PRIMER_INTERNAL_MAX_SELF_END': 8,
            'PRIMER_MIN_SIZE': 18,
            'PRIMER_MAX_SIZE': 25,
            'PRIMER_OPT_TM': 60.0,
            'PRIMER_MIN_TM': 57.0,
            'PRIMER_MAX_TM': 63.0,
            'PRIMER_MIN_GC': 20.0,
            'PRIMER_MAX_GC': 80.0,
            'PRIMER_MAX_POLY_X': 100,
            'PRIMER_INTERNAL_MAX_POLY_X': 100,
            'PRIMER_SALT_MONOVALENT': 50.0,
            'PRIMER_DNA_CONC': 50.0,
            'PRIMER_MAX_NS_ACCEPTED': 0,
            'PRIMER_MAX_SELF_ANY': 12,
            'PRIMER_MAX_SELF_END': 8,
            'PRIMER_PAIR_MAX_COMPL_ANY': 12,
            'PRIMER_PAIR_MAX_COMPL_END': 8,
            'PRIMER_PRODUCT_SIZE_RANGE': [
                [75, 100],
                [100, 125],
                [125, 150],
                [150, 175],
                [175, 200],
                [200, 225],
            ],
        }
        simulated_binding_res = simulatedbindings.design_primers(
            seq_args,
            global_args,
        )
        binding_res = bindings.design_primers(
            seq_args,
            global_args=global_args,
        )
        self._compare_results(
            binding_res,
            simulated_binding_res,
        )

    def test_file_based(self):
        test_file_roots = [
            'dv_conc_vs_dntp_conc',
            'long_seq',
            'p3-tmpl-mispriming',
            'primer_all_settingsfiles',
            'primer_check',
            'primer_end_pathology',
            'primer_first_base_index',
            'primer_gc_end',
            'primer_high_gc_load_set',
            'primer_high_tm_load_set',
            'primer_internal',
            'primer_must_overlap_point',
            'primer_must_use_th',
            'primer_num_best',
            'primer_ok_regions',
            'primer_overlap_junction',
            'primer_renewed_tasks',
            'primer_start_codon',
            'primer_task_th',
            'primer_task',
            'primer_thal_args',
            'primer_thal_max_seq_error',
            'primer_tm_lc_masking',
            'test_compl_error',
            'test_left_to_right_of_right',
        ]
        print()
        failures = []
        for fn_root in test_file_roots:
            base_fp = os.path.join(
                LOCAL_DIR,
                'input_files',
                fn_root,
            )
            input_fp = f'{base_fp}_input'
            log_path = os.path.join(
                LOCAL_DIR,
                'input_files_log',
            )
            os.makedirs(log_path, exist_ok=True)

            with open(input_fp) as input_fd:
                input_raw = input_fd.read()
            input_dicts = self._convert_boulder_input(input_raw)

            sys.stdout.write(f'->Testing file {fn_root:<40}\r')
            sys.stdout.flush()
            current_global_args = {}

            log_filepath_sim = os.path.join(log_path, f'{fn_root}_sim.log')
            if os.path.exists(log_filepath_sim):
                os.remove(log_filepath_sim)
            log_filepath_bind = os.path.join(log_path, f'{fn_root}_bind.log')
            if os.path.exists(log_filepath_bind):
                os.remove(log_filepath_bind)

            # NOTE: commented out for developement intentionally
            # sim_std_out_fp = 'sim_stdout.txt'
            # if os.path.exists(sim_std_out_fp):
            #     os.remove(sim_std_out_fp)
            # fd = open(sim_std_out_fp, 'wb')
            for global_args, seq_args, p3_args in input_dicts:
                current_global_args.clear()
                global_args['dump'] = 1
                test_id = str(seq_args.get('SEQUENCE_ID', ''))
                current_global_args.update(global_args)
                current_global_args['DO_LOG_SETTINGS'] = 1
                current_global_args['LOG_SETTINGS_PATH'] = log_filepath_sim

                simulated_binding_res = simulatedbindings.design_primers(
                    seq_args,
                    current_global_args,
                    # output_log=fd
                )
                current_global_args['LOG_SETTINGS_PATH'] = log_filepath_bind
                wrapper_error = simulated_binding_res.get('PRIMER_ERROR')
                if wrapper_error is not None:
                    print(wrapper_error)
                    with self.assertRaises((OSError, ValueError)):
                        binding_res = bindings.design_primers(
                            seq_args,
                            current_global_args,
                        )
                else:
                    try:
                        binding_res = bindings.design_primers(
                            seq_args,
                            current_global_args,
                        )
                    except (OSError, TypeError, ValueError):
                        if max([
                            x in p3_args.get('P3_COMMENT', '')
                            for x in ('complain', 'fail')
                        ]):
                            pass
                        raise
                    disagreements_str = self._compare_results(
                        binding_res,
                        simulated_binding_res,
                    )
                    if disagreements_str:
                        failures.append(
                            (fn_root, test_id, disagreements_str),
                        )
            # fd.close()
        print(' ' * 60, end='\r')
        if len(failures):
            err_msg = (
                'Failures occured during file testing:\n' +
                '\n'.join([
                    '->{}\t{}\n{}'.format(*f)
                    for f in failures
                ])
            )
            raise RuntimeError(err_msg)

    @unittest.skipIf(
        sys.platform == 'win32',
        'Windows does not support resource module',
    )
    def test_memory_leaks(self):
        sm = _get_mem_usage()
        run_count = 100
        for x in range(run_count):
            bindings.design_primers(
                {
                    'SEQUENCE_ID': 'MH1000',
                    'SEQUENCE_TEMPLATE': (
                        'GCTTGCATGCCTGCAGGTCGACTCTAGAGGATCCCCCTACATTTTAGCATCAG'
                        'TGAGTACAGCATGCTTACTGGAAGAGAGGGTCATGCAACAGATTAGGAGGTAA'
                        'GTTTGCAAAGGCAGGCTAAGGAGGAGACGCACTGAATGCCATGGTAAGAACTC'
                        'TGGACATAAAAATATTGGAAGTTGTTGAGCAAGTNAAAAAAATGTTTGGAAGT'
                        'GTTACTTTAGCAATGGCAAGAATGATAGTATGGAATAGATTGGCAGAATGAAG'
                        'GCAAAATGATTAGACATATTGCATTAAGGTAAAAAATGATAACTGAAGAATTA'
                        'TGTGCCACACTTATTAATAAGAAAGAATATGTGAACCTTGCAGATGTTTCCCT'
                        'CTAGTAG'
                    ),
                    'SEQUENCE_INCLUDED_REGION': [36, 342],
                },
                {
                    'PRIMER_OPT_SIZE': 20,
                    'PRIMER_PICK_INTERNAL_OLIGO': 1,
                    'PRIMER_INTERNAL_MAX_SELF_END': 8,
                    'PRIMER_MIN_SIZE': 18,
                    'PRIMER_MAX_SIZE': 25,
                    'PRIMER_OPT_TM': 60.0,
                    'PRIMER_MIN_TM': 57.0,
                    'PRIMER_MAX_TM': 63.0,
                    'PRIMER_MIN_GC': 20.0,
                    'PRIMER_MAX_GC': 80.0,
                    'PRIMER_MAX_POLY_X': 100,
                    'PRIMER_INTERNAL_MAX_POLY_X': 100,
                    'PRIMER_SALT_MONOVALENT': 50.0,
                    'PRIMER_DNA_CONC': 50.0,
                    'PRIMER_MAX_NS_ACCEPTED': 0,
                    'PRIMER_MAX_SELF_ANY': 12,
                    'PRIMER_MAX_SELF_END': 8,
                    'PRIMER_PAIR_MAX_COMPL_ANY': 12,
                    'PRIMER_PAIR_MAX_COMPL_END': 8,
                    'PRIMER_PRODUCT_SIZE_RANGE': [
                        [75, 100],
                        [100, 125],
                        [125, 150],
                        [150, 175],
                        [175, 200],
                        [200, 225],
                    ],
                },
            )
        sleep(0.1)  # Pause for any GC
        em = _get_mem_usage()
        print(
            f'\n\tMemory usage before {run_count} runs of design_primers: {sm}',
        )
        print(
            f'\tMemory usage after {run_count} runs of design_primers: {em}',
        )
        print(f'\t\t\t\t\tDifference: \t {em - sm}')

        # NOTE: MacOS has a different allocation strategy than Linux
        # Periodically revisit this. 2023.01.09
        delta_bytes_limit = 1000 if sys.platform == 'linux' else 10000

        if em - sm > delta_bytes_limit:
            raise AssertionError(
                f'Memory usage increase after {run_count} runs of \n\t'
                f'design_primers > {delta_bytes_limit} bytes -- potential \n\t'
                f'memory leak (mem increase: {em - sm})',
            )

    def test_misprime_lib_mishyb_lib(self):
        '''Test that the misprime library and mishyb library arguments
        successfully run through the bindings.

        '''
        seq_args = {
            'SEQUENCE_ID': 'MH1000',
            'SEQUENCE_TEMPLATE': (
                'GCTTGCATGCCTGCAGGTCGACTCTAGAGGATCCCCCTACATTTT'
                'AGCATCAGTGAGTACAGCATGCTTACTGGAAGAGAGGGTCATGCA'
                'ACAGATTAGGAGGTAAGTTTGCAAAGGCAGGCTAAGGAGGAGACG'
                'CACTGAATGCCATGGTAAGAACTCTGGACATAAAAATATTGGAAG'
                'TTGTTGAGCAAGTNAAAAAAATGTTTGGAAGTGTTACTTTAGCAA'
                'TGGCAAGAATGATAGTATGGAATAGATTGGCAGAATGAAGGCAAA'
                'ATGATTAGACATATTGCATTAAGGTAAAAAATGATAACTGAAGAA'
                'TTATGTGCCACACTTATTAATAAGAAAGAATATGTGAACCTTGCA'
                'GATGTTTCCCTCTAGTAG'
            ),
            'SEQUENCE_INCLUDED_REGION': [36, 342],
        }
        global_args = {
            'PRIMER_OPT_SIZE': 20,
            'PRIMER_PICK_INTERNAL_OLIGO': 1,
            'PRIMER_INTERNAL_MAX_SELF_END': 8,
            'PRIMER_MIN_SIZE': 18,
            'PRIMER_MAX_SIZE': 25,
            'PRIMER_OPT_TM': 60.0,
            'PRIMER_MIN_TM': 57.0,
            'PRIMER_MAX_TM': 63.0,
            'PRIMER_MIN_GC': 20.0,
            'PRIMER_MAX_GC': 80.0,
            'PRIMER_MAX_POLY_X': 100,
            'PRIMER_INTERNAL_MAX_POLY_X': 100,
            'PRIMER_SALT_MONOVALENT': 50.0,
            'PRIMER_DNA_CONC': 50.0,
            'PRIMER_MAX_NS_ACCEPTED': 0,
            'PRIMER_MAX_SELF_ANY': 12,
            'PRIMER_MAX_SELF_END': 8,
            'PRIMER_PAIR_MAX_COMPL_ANY': 12,
            'PRIMER_PAIR_MAX_COMPL_END': 8,
            'PRIMER_PRODUCT_SIZE_RANGE': [
                [75, 100], [100, 125], [125, 150],
                [150, 175], [175, 200], [200, 225],
            ],
        }

        result = bindings.design_primers(
            seq_args=seq_args,
            global_args=global_args,
            misprime_lib={
                'SEQ1': 'CACCATGGAGCTCCTGATATTAAAGGCGAATGCCATT',
            },
        )
        self.assertEqual(result['PRIMER_PAIR_NUM_RETURNED'], 5)
        self.assertEqual(result['PRIMER_LEFT_0'], [46, 21])

        result = bindings.design_primers(
            seq_args=seq_args,
            global_args=global_args,
            mishyb_lib={
                'SEQ1': 'TTATGTGCCACACTTATTAATAAGAAAGAATATGTGAACCTTGCA',
            },
        )
        self.assertEqual(result['PRIMER_PAIR_NUM_RETURNED'], 5)
        self.assertEqual(result['PRIMER_LEFT_0'], [46, 21])

        bindings.design_primers(
            seq_args=seq_args,
            global_args=global_args,
            misprime_lib={
                'SEQ1': 'CACCATGGAGCTCCTGATATTAAAGGCGAATGCCATT',
            },
            mishyb_lib={
                'SEQ1': 'TTATGTGCCACACTTATTAATAAGAAAGAATATGTGAACCTTGCA',
            },
        )
        self.assertEqual(result['PRIMER_PAIR_NUM_RETURNED'], 5)
        self.assertEqual(result['PRIMER_LEFT_0'], [46, 21])


def suite():
    suite = unittest.TestSuite()
    suite.addTest(TestDesignBindings())
    return suite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    test_suite = suite()
    runner.run(test_suite)



# Copyright (C) 2023. Ben Pruitt & Nick Conway;
# See LICENSE for full GPLv2 license.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
'''
tests.test_ntthal_io
~~~~~~~~~~~~~~~~~~~~~~~

Test to confirm primer3-py updates to the thal.c library result in the same
output given the input, `thal_input` and expected output, `thal_output` file
provided by the primer3 source library
'''
import io
import os
import os.path as op
import subprocess
import unittest
from typing import List

TEST_DIR = op.dirname(op.realpath(__file__))
PACKAGE_DIR = op.dirname(TEST_DIR)
PRIMER3_DIR = op.join(PACKAGE_DIR, 'primer3')
INPUT_DIR = op.join(TEST_DIR, 'input_files')
LIBPRIMER3_PATH = op.join(PRIMER3_DIR, 'src', 'libprimer3')
NTTHAL_PATH = op.join(LIBPRIMER3_PATH, 'ntthal')


def split_args_line_to_list(line: str) -> List[str]:
    '''
    Args:
        line: a line of arguments to split by space.  Filter out empty strings
            if there are extra spaces between arguments

    Returns:
        List of strings of split arguments
    '''
    args_line_list = line.split(' ')
    # filter out empty strings to prevent parsing errors
    return [x for x in args_line_list if x != '']


def run_ntthal(line: str, stderr_fd: io.BufferedWriter) -> str:
    '''
    Run ntthal given the argument line

    Args:
        line: A line of arguments for the ntthal command
        stderr_fd: Writter to send stderr to such that it does not go to stdout

    Returns:
        On success, the string of lines of the process stdout printing.
        On failure an empty string
    '''
    args_line_list = split_args_line_to_list(line)
    args = [NTTHAL_PATH] + args_line_list
    try:
        out_b = subprocess.check_output(
            args,
            stderr=stderr_fd,
            env=os.environ,
        )
        # Need to trim trailing white space as thal_output is trimmed by
        # pre-commit
        return out_b.decode('utf8')

    except subprocess.CalledProcessError:
        # NOTE: Intentionally commented out for debugging purposes
        # print(line)
        return ''


class TestLowLevelBindings(unittest.TestCase):
    def test_confirm_ntthal_binary(self):
        '''Confirm updates to thal.c do not change expected output

        NOTE: thal_input has several argument lines that are INTENTIONAL
        failures. Failure argument lines result in no output in thal_output and
        therefore there are fewer output line sets that input argument lines
        '''
        compute_str_list: List[str] = []
        with open(op.join(INPUT_DIR, 'thal_input'), 'r') as fd:
            input_str = fd.read()
        with open(op.join(INPUT_DIR, 'thal_output'), 'r') as fd:
            output_str = fd.read()
        with open(os.devnull, 'wb') as dev_null:
            for line in input_str.split('\n'):
                compute_str_list.append(run_ntthal(line, dev_null))
        # Replace line endings for windows compat
        compute_str = ''.join(compute_str_list).replace('\r\n', '\n')
        self.assertEqual(compute_str, output_str)
        # NOTE: Intentionally commented out for debugging purposes
        # if compute_str != output_str:
        #     # print(compute_str)
        #     with open('thal_input_compute_str', 'w') as fd:
        #             fd.write(compute_str)
        #     raise ValueError('')
